/*

NDO.java

*/

package cs141b.gndo;

import java.io.DataOutputStream;
import java.io.DataInputStream;
import java.io.IOException;
import java.util.Vector;
import java.util.Enumeration;
import info.net.Place;
import info.net.MailDaemon;
import info.net.Receivebox;
import info.net.Sendbox;
import info.net.Broadcastbox;

public abstract class NDO extends SharedContainer {
  private MailDaemon _md;
  private Receivebox _inbox;
  private Sendbox _outbox; //Rebind at will
  private Broadcastbox _broadcastBox; //Broadcast box - updated every time the NDO is
  private PathList _lockList; //List of paths
  private InboxWatcher _inboxWatcher; //Thread that is watching the inbox
  private Vector _workerList; //List of places
  private StateInt _leaveCounter; //A variable that counts leaves

  /** Creating constructor */
  public NDO() {
    super();
    _ndo = this;

    try {
      _md = new MailDaemon();
      _inbox = new Receivebox(_md, "inbox");
      //      _joinbox = new Receivebox(_md, "join");
      _outbox = new Sendbox(_md, "outbox");
      _broadcastBox = new Broadcastbox(_md, "broadcastbox");
    }
    catch(info.net.MailDaemonException e) {
      throw new RuntimeException(e.toString());
    }

    _lockList = null;
    _workerList = new Vector();
    _workerList.addElement(where());

    makeFields();

    unlockAll();

    spawnThreads();
  }

  /** Joining constructor */
  public NDO(Place place) {
    super();
    _ndo = this;

    try {
      _md = new MailDaemon();
      _inbox = new Receivebox(_md, "inbox");
      //      _joinbox = new Receivebox(_md, "join");
      _outbox = new Sendbox(_md, "outbox");
      _broadcastBox = new Broadcastbox(_md, "broadcastbox");
    }
    catch(info.net.MailDaemonException e) {
      throw new RuntimeException(e.toString());
    }

    _lockList = null;
    _workerList = null;

    makeFields();

    sendMessage(place, new WantLockMessage(where(),
                                           new Path(this)));

    GiveLockMessage glMessage = (GiveLockMessage)_inbox.receive();
    Assert.isNotNull(glMessage);

    glMessage.which().update(this);

    Assert.isTrue(glMessage.which().equals(new Path(this)));

    glMessage.update(this);

    getObject(glMessage.wantLockQueue());

    Assert.isTrue(locked());
    Assert.isTrue(haveObject());
    Assert.isFalse(modified());
    Assert.isNull(lastGaveLock());
    Assert.isNull(bestGuessOwner());

    initialSetupPropogateChildren(glMessage.from());

    Assert.isNotNull(_workerList);
    _workerList.addElement(where());
    incCurrentState(); // We modified the NDO

    JoinMessage jMessage = new JoinMessage(where(),
                                           new DataContainingMessage(where(),
                                                                     new Path(this)));

    for(int i=0;i<_workerList.size();++i) {
      Place curDest = (Place)_workerList.elementAt(i);
      
      if (!curDest.equals(where())) {
        sendMessage(curDest, jMessage);
      }
    }

    unlock();
    spawnThreads();
  }

  /** This must be overridden so that it makes the basic fields in this object */
  public abstract void makeFields();

  /** This method must be called to gracefully disconnect from a shared object */
  public /*synchronized*/ void leave() {
    //Get lock on NDO
    startLock();
    lock();
    try {
      completeLock();
    }
    catch(FieldDeletedException e) {
      Assert.error("Can't delete the NDO!");
    }

    //Remove ourselves from the workerList
    
    Assert.isNotNull(_workerList);
    Assert.isFalse(_workerList.isEmpty());

    _workerList.removeElement(where());
    incCurrentState(); // We modified the NDO

    if (!_workerList.isEmpty()) {
      //Pick recipient

      Place recipient = (Place)_workerList.elementAt(0);

      //Send off all fields we own with a TakeThisMessage, or a GiveLockMessage if the
      //recipient is the first person in the wantLockQueue for that field

      // Push our children
      pushChildren(recipient);

      //Send messages to all other workers asking them to tell us all the fields for which
      //we are their best guess. This message should also be an update on the NDO, and
      //should set the best guess for the NDO to the recipient
      //Before we send the messages, make sure we know how many we're sending

      _leaveCounter = new StateInt(_workerList.size());

      LeaveStartedMessage lsMessage = new LeaveStartedMessage(where(),
                                                              new DataContainingMessage(recipient,
                                                                                        new Path(this)));

      for(int i=0;i<_workerList.size();++i) {
        Place curDest = (Place)_workerList.elementAt(i);

        Assert.isFalse(curDest.equals(where()));

        sendMessage(curDest, lsMessage);
      }

      //Receive a response from every worker as to which fields we are their best guess for
      //Send each worker a response saying who their best guess should be for that field (our best guess)
      //This is done in the message handler

      //Receive a response from every worker acknowleding our departure
      //(Once we receive a message from someone acknowledging our leave, they should _never_
      //send us anything more) The response receive is done in the message handler,
      //but we need to block here until it happens. We do this by using the NDO variable
      //leaveCounter

      _leaveCounter.lock();
      _leaveCounter.waitForValue(0);
      _leaveCounter.unlock();

      //Send off the NDO with a TakeThisMessage, or a GiveLockMessage if the
      //recipient is the first person in the wantLockQueue for the NDO

      Assert.isNotNull(wantLockQueue());
      if (!wantLockQueue().isEmpty() && recipient.equals(wantLockQueue().elementAt(0))) {
        wantLockQueue().removeElementAt(0);

        sendMessage(recipient,
                    new GiveLockMessage(where(),
                                        wantLockQueue(),
                                        new Path(this)));
      } else {
        Assert.isTrue(_varState.quickValue() == NORMAL);
        Assert.isTrue(haveObject());
        Assert.isTrue(locked());

        sendMessage(recipient,
                    new TakeThisMessage(where(),
                                        wantLockQueue(),
                                        new Path(this)));

        //Setup state things, even though we really don't need to

//         _lastGaveLock = recipient;
//         _bestGuessOwner = _lastGaveLock;
//         _haveObject = false;
//         setLocked(false);
      }
    }

    //Cause thread to exitKill threads

     _outbox.bind(where());
     _outbox.send(new KillMessage(where()));
     try {
       _inboxWatcher.join();
     }
     catch(InterruptedException e) {
       Assert.error("InterruptedException: "+e);
     }
  }

  public Place where() {
    return _inbox.addr();
  }

  /** This can be used to send messages */
  void sendMessage(Place where, Message message) {
    synchronized(_outbox) {
      Assert.isFalse(where.equals(this.where()));

      _outbox.bind(where);
      _outbox.send(message);
    }
  }

  /** This can be used to broadcast messages */
  synchronized void broadcastMessage(Message message) {
    /*
    System.out.println("NDO.broadcastMessage("+message.getClass().getName()+")");
    boolean first=false;
    System.out.print("[");
    Enumeration e=_broadcastBox.destination();
    while(e.hasMoreElements()) {
      Place p=(Place)e.nextElement();
      if (first) System.out.print(" ");
      System.out.print(p);
      first = false;
    }
    System.out.println("]");
    */
    _broadcastBox.send(message);
  }

  void registerLock(SharedVar sv) {
    Assert.isNotNull(_lockList);

    synchronized(_lockList) {
      _lockList.add(new Path(sv));
    }
  }

  Vector workerList() {
    return _workerList;
  }

  public synchronized void readData(DataInputStream dis) throws IOException {
    super.readData(dis);

    Assert.isNotNull(_broadcastBox);
    _broadcastBox.removeAll();

    int len = dis.readInt();
    _workerList = new Vector(len);
    for(int i=0;i<len;++i) {
      Place place = new Place();

      place.readData(dis);

      _workerList.addElement(place);
      if (!place.equals(where())) {
        _broadcastBox.add(place);
      }
    }
    
  }

  public synchronized void writeData(DataOutputStream dos) throws IOException {
    super.writeData(dos);

    dos.writeInt(_workerList.size());
    for(int i=0;i<_workerList.size();++i) {
      ((Place)_workerList.elementAt(i)).writeData(dos);
    }
  }

  public void startLock() {
    _lockList = new PathList();
  }

  public final void completeLock() throws FieldDeletedException {
    Assert.isNotNull(_lockList);

    synchronized(_lockList) {
      for(int i=0;i<_lockList.size();++i) {
        Path curPath = (Path)_lockList.elementAt(i);
        SharedVar curVar = curPath.leaf();
        
        Assert.isFalse(curVar.locked());
        if (curVar.haveObject()) {
          curVar.getObject(null);
        } else {
          curVar._varState.lock();
          Assert.isTrue(curVar._varState.value() == NORMAL);
          curVar._varState.setValue(WAITING_FOR_LOCK);
          curVar._varState.unlock();
          
          sendMessage(curVar.bestGuessOwner(), new WantLockMessage(where(), curPath));
          
          curVar._varState.lock();
          curVar._varState.waitForNotValue(SharedVar.WAITING_FOR_LOCK);
          int curVarState = curVar._varState.value();
          curVar._varState.unlock();
          
          if (curVarState != SharedVar.NORMAL) {
            Assert.isTrue(curVarState == SharedVar.DELETED);
            
            for(int j=0;j<i;++j) {
              SharedVar curLockedVar = ((Path)_lockList.elementAt(i)).leaf();
              
              Assert.isTrue(curLockedVar.locked());
              
              curLockedVar.unlock();
            }
            
            throw new FieldDeletedException();
          }
        }
      }
    }
  }

  void spawnThreads() {
    _inboxWatcher = new InboxWatcher(this);
    _inboxWatcher.start();
//     _joinboxWatcher = new JoinboxWatcher(this);
//     _joinboxWatcher.start();
  }

  void watchInbox() {
    boolean done = false;

    Assert.isNotNull(_inbox);

    while(!done) {
      Message message = (Message)_inbox.receive();
      Assert.isNotNull(message);

      if (message instanceof VarMessage) {
        VarMessage vMessage = (VarMessage)message;

        vMessage.which().update(this);

        if (vMessage.which().valid()) {
          // Dispatch message

          SharedVar toWhom = vMessage.which().leaf();

          toWhom.handleMessage(vMessage);
        } else {
          if (vMessage.which().whyInvalid() == Path.FIELD_DELETED) {
            // If the field was deleted, respond with an update on the container
            // that got changed to make the path invalid

            sendMessage(vMessage.from(), 
                        new DataContainingMessage(where(),
                                                  new Path(vMessage.which().getLastValid())));
          }
        }
      } else if (message instanceof LeaveStartedMessage) {
        LeaveStartedMessage lsMessage = (LeaveStartedMessage)message;

        lsMessage.ndoUpdate().which().update(this);
        Assert.isTrue(lsMessage.ndoUpdate().which().leaf() == this);
        Assert.isTrue(lsMessage.ndoUpdate().stateID() > currentState());
        lsMessage.ndoUpdate().update(this);

        PathList listOfGaveLock = new PathList();
        PathList listOfBestGuesses = new PathList();

        findPointersTo(listOfGaveLock, listOfBestGuesses,
                       lsMessage.from());

        if (listOfGaveLock.isEmpty() && listOfBestGuesses.isEmpty()) {
          sendMessage(lsMessage.from(), new LeaveAcknowledgedMessage(where()));
        } else {
          sendMessage(lsMessage.from(), new RequestBetterGuessesMessage(where(),
                                                                        listOfGaveLock,
                                                                        listOfBestGuesses));
        }
      } else if (message instanceof RequestBetterGuessesMessage) {
        RequestBetterGuessesMessage rbgMessage = (RequestBetterGuessesMessage)message;
        PathList whichFieldsGaveLock = rbgMessage.whichFieldsGaveLock();
        PathList whichFieldsBestGuesses = rbgMessage.whichFieldsBestGuesses();
        Vector newGaveLock = new Vector(whichFieldsGaveLock.size());
        Vector newBestGuesses = new Vector(whichFieldsBestGuesses.size());

        for(int i=0;i<whichFieldsGaveLock.size();++i) {
          Path curPath = (Path)whichFieldsGaveLock.elementAt(i);
          curPath.update(this);

          newGaveLock.addElement(curPath.leaf().lastGaveLock());
        }
        for(int i=0;i<whichFieldsBestGuesses.size();++i) {
          Path curPath = (Path)whichFieldsBestGuesses.elementAt(i);
          curPath.update(this);

          newBestGuesses.addElement(curPath.leaf().bestGuessOwner());
        }

        Assert.isTrue(whichFieldsGaveLock.size() == newGaveLock.size());
        Assert.isTrue(whichFieldsBestGuesses.size() == newBestGuesses.size());

        sendMessage(rbgMessage.from(),
                    new NewOwnerMessage(where(),
                                        whichFieldsGaveLock,
                                        newGaveLock,
                                        whichFieldsBestGuesses,
                                        newBestGuesses));
      } else if (message instanceof NewOwnerMessage) {
        NewOwnerMessage noMessage = (NewOwnerMessage)message;
        PathList whichFieldsGaveLock = noMessage.whichFieldsGaveLock();
        PathList whichFieldsBestGuesses = noMessage.whichFieldsBestGuesses();
        Vector newGaveLock = noMessage.newGaveLock();
        Vector newBestGuesses = noMessage.newBestGuesses();

        Assert.isTrue(whichFieldsGaveLock.size() == newGaveLock.size());
        Assert.isTrue(whichFieldsBestGuesses.size() == newBestGuesses.size());

        for(int i=0;i<whichFieldsGaveLock.size();++i) {
          Path curPath = (Path)whichFieldsGaveLock.elementAt(i);
          curPath.update(this);

          curPath.leaf().setLastGaveLock((Place)newGaveLock.elementAt(i));
        }
        for(int i=0;i<whichFieldsBestGuesses.size();++i) {
          Path curPath = (Path)whichFieldsBestGuesses.elementAt(i);
          curPath.update(this);

          curPath.leaf().setBestGuessOwner((Place)newBestGuesses.elementAt(i));
        }

        sendMessage(noMessage.from(), new LeaveAcknowledgedMessage(where()));

      } else if (message instanceof LeaveAcknowledgedMessage) {
        LeaveAcknowledgedMessage laMessage = (LeaveAcknowledgedMessage)message;

        _leaveCounter.lock();
        Assert.isTrue(_leaveCounter.value() > 0);
        _leaveCounter.setValue(_leaveCounter.value()-1);
        _leaveCounter.unlock();
      } else if (message instanceof JoinMessage) {
        JoinMessage jMessage = (JoinMessage)message;

        jMessage.ndoUpdate().which().update(this);
        Assert.isTrue(jMessage.ndoUpdate().which().leaf() == this);
        Assert.isTrue(jMessage.ndoUpdate().stateID() > currentState());

        jMessage.ndoUpdate().update(this);
      } else if (message instanceof KillMessage) {
        done = true;
      }
    }
  }

  /*
  void watchJoinbox() {
    boolean done = false;

    Assert.isNotNull(_joinbox);

    while(!done) {
      JoinMessage jMessage = (JoinMessage)_joinbox.receive();
      Assert.isNotNull(jMessage);
      Assert.isNotNull(_inbox);
      Assert.isNotNull(_workerList);

      _workerList.addElement(jMessage.where());

      sendMessage(jMessage.where(), new BootstrapMessage(where(),
                                                         _workerList));
    }
  }
  */
}
