/*

SharedVar.java

*/

package cs141b.gndo;

import java.io.DataOutputStream;
import java.io.DataInputStream;
import java.io.IOException;
import info.net.Place;
import java.util.Vector;
import java.util.Observable;

public abstract class SharedVar extends Observable {
  static final int NORMAL = 0;
  static final int WAITING_FOR_LOCK = 1;
  static final int DELETED = 2;
  static final int FAULT_RECOVERY = 3;

  SharedContainer _parent; // Always valid
  NDO _ndo; // Always valid
  int _myIndex; // Always valid, unless this is an NDO
  private boolean _haveObject; // true when the local computer has charge of this SharedVar
  private boolean _locked; // true when the local computer has a lock on this SharedVar (implies _haveObject)
  private boolean _modified; // true when this object has locked the SharedVar and made changes to it
  private Place _lastGaveLock; // null when _haveObject and _haveObject has never been true. Once
                               // we've had the object once, this will be non-null when we don't have it,
                               // and it points to who we last gave the lock to
  private Place _bestGuessOwner; // null when _haveObject is true. Once the object has been initialized fully
                                 // is non-null. This will be non-null when we don't have it,
                                 // and it will point to somebody who is at least as good as _lastGaveLock
                                 // to ask for the object
  private Vector _wantLockQueue; // Always non-null. Should be empty when we don't have the lock
                                 // When we do have the lock, holds a list of WantLockMessages corresponding
                                 // to who wants it next
  StateInt _varState; // One of the above constants
  private long _currentState; // This is the current state of this object. This is used to
                      // order the states of object when messages come in.
  Place _readDataFrom; // Who sent us the data we're reading in this call to readData
  StateInt _recoverMessageCount; // How many recover messages we've received

  public SharedVar() {
    _parent = null;
    _ndo = null;
    _myIndex = -1;
    _haveObject = true;
    _locked = true;
    _modified = false;
    _lastGaveLock = null;
    _bestGuessOwner = null;
    _wantLockQueue = new Vector();
    _varState = new StateInt(NORMAL);
    _currentState = 0;
    _recoverMessageCount = null;
  }

  /**The constructor
   @param parent Specify which item (NDO or other SharedContainer) this will be a field of
   @see cs141b.gndo.SharedContainer
   @see cs141b.gndo.NDO*/
  public SharedVar(SharedContainer parent) {
    _parent = parent;
    _ndo = _parent._ndo;
    Assert.isNotNull(_ndo);
    _myIndex = -1;
    _haveObject = true;
    _locked = true;
    _modified = false;
    _lastGaveLock = null;
    _bestGuessOwner = null;
    _wantLockQueue = new Vector();
    _parent.addField(this);
    _varState = new StateInt(NORMAL);
    _currentState = 0;
    _recoverMessageCount = null;
  }

  /**Returns true when this object is locked, and may be modified. DO NOT modify the object
    if this returns false!*/
  public synchronized final boolean locked() {
    return _locked;
  }

  synchronized void setLocked(boolean locked) {
    _locked = locked;
  }
  
  /**Returns true when this object HAS BEEN modified.*/
  public synchronized final boolean modified() {
    return _modified;
  }

  /**Call this when you modify an object after having it locked.*/
  public synchronized final void setModified() {
    _modified = true;
  }

  synchronized final void setUnmodified() {
    _modified = false;
  }

  synchronized boolean haveObject() {
    return _haveObject;
  }

  synchronized void setHaveObject(boolean haveObject) {
    _haveObject = haveObject;
  }

  synchronized Place lastGaveLock() {
    return _lastGaveLock;
  }

  synchronized void setLastGaveLock(Place lastGaveLock) {
    _lastGaveLock = lastGaveLock;
  }

  synchronized Place bestGuessOwner() {
    return _bestGuessOwner;
  }

  synchronized void setBestGuessOwner(Place bestGuessOwner) {
    _bestGuessOwner = bestGuessOwner;
  }

  synchronized Vector wantLockQueue() {
    return _wantLockQueue;
  }

  /**Get the current state of the object*/
  public synchronized final long currentState() {
    return _currentState;
  }

  /**Set the current state of the object*/
  synchronized final void setCurrentState(long currentState) {
    Assert.isTrue(currentState > _currentState);

    _currentState = currentState;

    setChanged();
    notifyObservers();
  }

  /**Increment the current state of the object*/
  synchronized final void incCurrentState() {
    ++_currentState;

    setChanged();
    notifyObservers();
  }

  synchronized void delete() {
    _myIndex = -1;
    _locked = false;
    _lastGaveLock = null;
    _bestGuessOwner = null;
    _varState.lock();
    _varState.setValue(DELETED);
    _varState.unlock();
    _currentState = -1;
  }

  /**Call this after calling NDO.startLock to indicate that you want to lock this field.
    You cannot start modifying fields until after calling NDO.completeLock*/
  public synchronized final void lock() {
    Assert.isFalse(_locked);
    //    Assert.isNotNull(_ndo); //We must not be the NDO
    //    Assert.isNotNull(_parent);

    _ndo.registerLock(this);
  }

  /**Call this after you are done modifying this field. Do not call between calls to
    NDO.startLock and NDO.completeLock.*/
  public synchronized final void unlock() {
    Assert.isTrue(_locked);
    Assert.isTrue(_haveObject);
    Assert.isNotNull(_ndo);
    Assert.isNotNull(_wantLockQueue);

    if (_modified) {
      Assert.isTrue(_currentState < Long.MAX_VALUE);
      incCurrentState();

      _ndo.broadcastMessage(new DataContainingMessage(_ndo.where(),
                                                      new Path(this)));
    }

    _locked = false;

    // See if somebody wants this object

    if (!_wantLockQueue.isEmpty()) {
      // Somebody is waiting for this, give it to them

      WantLockMessage wlMessage = (WantLockMessage)_wantLockQueue.elementAt(0);
      _wantLockQueue.removeElementAt(0);

      _ndo.sendMessage(wlMessage.from(), 
                       new GiveLockMessage(_ndo.where(),
                                           _wantLockQueue,
                                           wlMessage.which())); 
      _lastGaveLock = wlMessage.from();
      _bestGuessOwner = _lastGaveLock;

      _haveObject = false;

      Assert.isTrue(_varState.quickValue() == NORMAL);
    } else {
      // Sniff... nobody wants us!
    }
  }

  /**Call this to request an update of the current state of the object, without locking it*/
  public synchronized void getUpdate() {
    if (!_haveObject) {
System.out.println("Sending message to "+_bestGuessOwner+" asking for update on "+new Path(this)+" current state "+_currentState);
      _ndo.sendMessage(_bestGuessOwner,
                       new WantUpdateMessage(_ndo.where(),
                                             new Path(this),
                                             _currentState));
    }
  }
  
  public synchronized void lockAll() {
    lock();
  }

  public synchronized void unlockAll() {
    if (_locked) unlock();
  }

  // Functions to be implemented by subclass
  
  /**Read in all of the fields not derived from SharedVar from an outside stream*/
  public abstract synchronized void readData(DataInputStream dis) throws IOException;
    
  /**Write out all of the fields not derived from SharedVar to an outside stream*/
  public abstract synchronized void writeData(DataOutputStream dos) throws IOException;

  // Package functions

  final synchronized void handleMessage(VarMessage message) {

    System.out.println("message=="+message.getClass().getName()+" to a "+getClass().getName()+" at "+message.which()+" from "+message.from());

    if (message instanceof WantLockMessage) {
      WantLockMessage wlMessage = (WantLockMessage)message;

      if (_haveObject) {
        // We have the object, so we can send it off

        if (_locked) {
          Assert.isNotNull(_wantLockQueue);

          _wantLockQueue.addElement(wlMessage);
        } else {
          Assert.isNotNull(_wantLockQueue);
          Assert.isTrue(_wantLockQueue.isEmpty());

          _ndo.sendMessage(wlMessage.from(), 
                           new GiveLockMessage(_ndo.where(),
                                               _wantLockQueue,
                                               wlMessage.which())); 
          _lastGaveLock = wlMessage.from();
          _bestGuessOwner = _lastGaveLock;
          _wantLockQueue.removeAllElements();
          _haveObject = false;
        }
      } else {
        // We don't have the object, forward the message

        Assert.isNotNull(_wantLockQueue);
        Assert.isTrue(_wantLockQueue.isEmpty());
        Assert.isNotNull(_bestGuessOwner);

        _ndo.sendMessage(_bestGuessOwner, wlMessage);
      }
    } else if (message instanceof WantUpdateMessage) {
      WantUpdateMessage wuMessage = (WantUpdateMessage)message;

      if (_haveObject) {
        // We have the object, so we may be able to update the person

        if (!_locked && _currentState > wuMessage.currentState() &&
            !_ndo.where().equals(wuMessage.from())) {
          _ndo.sendMessage(wuMessage.from(),
                           new DataContainingMessage(_ndo.where(),
                                                     wuMessage.which()));
        }
      } else {
        Assert.isNotNull(_bestGuessOwner);

        _ndo.sendMessage(_bestGuessOwner, wuMessage);
      }
    } else if (message instanceof GiveLockMessage) {
      GiveLockMessage glMessage = (GiveLockMessage)message;

      if (glMessage.stateID() > _currentState) {
        glMessage.update(this);
      }

      getObject(glMessage.wantLockQueue());
    } else if (message instanceof TakeThisMessage) {
      TakeThisMessage ttMessage = (TakeThisMessage)message;

      if (ttMessage.stateID() > _currentState) {
        ttMessage.update(this);
      }

      takeObject(ttMessage.wantLockQueue());
    } else if (message instanceof DataContainingMessage) {
      DataContainingMessage dcMessage = (DataContainingMessage)message;

      if (dcMessage.stateID() > _currentState) {
        dcMessage.update(this);
      }
    }
  }

  final synchronized void getObject(Vector wantLockQueue) {
    // Handle stuff to give us locks
    
    _haveObject = true;
    _locked = true;
    _modified = false;
    _lastGaveLock = null;
    _bestGuessOwner = null;
    if (wantLockQueue != null) {
      _wantLockQueue = wantLockQueue;
    }

    _varState.lock();
    _varState.setValue(NORMAL);
    _varState.unlock();
  }

  final synchronized void takeObject(Vector wantLockQueue) {
    Assert.isTrue(_wantLockQueue.isEmpty());

    // Handle stuff to take ownership, but not to take the lock

    if (!wantLockQueue.isEmpty()) {
      // Somebody is waiting for this, give it to them

      WantLockMessage wlMessage = (WantLockMessage)_wantLockQueue.elementAt(0);
      _wantLockQueue.removeElementAt(0);

      Assert.isFalse(_ndo.where().equals(wlMessage.from()));

      _ndo.sendMessage(wlMessage.from(),
                       new GiveLockMessage(_ndo.where(),
                                           wantLockQueue,
                                           wlMessage.which()));

      _lastGaveLock = wlMessage.from();
      _bestGuessOwner = _lastGaveLock;

      _haveObject = false;
      Assert.isFalse(_locked);      
    } else {
      //The wantLockQueue is empty

      _haveObject = true;
      Assert.isFalse(_locked);
      _lastGaveLock = null;
      _bestGuessOwner = null;
      Assert.isTrue(_varState.quickValue() == NORMAL);
    }
  }

  synchronized void pushObject(Place where) {
    Assert.isNotNull(where);
    Assert.isFalse(_locked);
    Assert.isTrue(_varState.quickValue() == NORMAL);

    if (_haveObject) {
      Assert.isNotNull(_wantLockQueue);
      Assert.isTrue(_wantLockQueue.isEmpty());

      _ndo.sendMessage(where, new TakeThisMessage(_ndo.where(),
                                                  _wantLockQueue,
                                                  new Path(this)));

      _lastGaveLock = where;
      _bestGuessOwner = _lastGaveLock;
      _haveObject = false;
    }
  }

  synchronized void findPointersTo(PathList resultBinGaveLock, PathList resultBinBestGuesses,
                      Place findLocation) {
    if (findLocation.equals(_lastGaveLock)) {
      Assert.isNotNull(resultBinGaveLock);
      
      resultBinGaveLock.add(new Path(this));
    }
    if (findLocation.equals(_bestGuessOwner)) {
      Assert.isNotNull(resultBinBestGuesses);
      
      resultBinBestGuesses.add(new Path(this));
    }
  }

  synchronized void initialSetupPropogate(Place where) {
    _locked = false;
    _haveObject = false;
    _modified = false;
    _lastGaveLock = null;
    _bestGuessOwner = where;

    Assert.isNotNull(_wantLockQueue);
    Assert.isTrue(_wantLockQueue.isEmpty());
    Assert.isTrue(_varState.quickValue() == NORMAL);

    getUpdate();
  }

  synchronized void faultRecover(Place whereFaulted) {
    Assert.isNotNull(_ndo);
    Vector workerListImage = _ndo.workerList();
    Assert.isNotNull(workerListImage);

    _recoverMessageCount = new StateInt(workerListImage.size());
    _varState.lock();
    _varState.setValue(FAULT_RECOVERY);
    _varState.unlock();

    FaultDetectedMessage fdMessage = new FaultDetectedMessage(_ndo.where(),
                                                              new Path(this),
                                                              whereFaulted);

    for(int i=0;i<workerListImage.size();++i) {
      Place curWorker = (Place)workerListImage.elementAt(i);

      if (!curWorker.equals(_ndo.where()) && !curWorker.equals(whereFaulted)) {
        _ndo.sendMessage(curWorker, fdMessage);
      } else {
        _recoverMessageCount.lock();
        _recoverMessageCount.setValue(_recoverMessageCount.value()-1);
        _recoverMessageCount.unlock();
      }
    }

    _recoverMessageCount.lock();
    _recoverMessageCount.waitForValue(0);
    _recoverMessageCount.unlock();

    _varState.lock();
    int varStateValue = _varState.value();
    _varState.unlock();

    if (varStateValue == FAULT_RECOVERY) {
      // Variable has already been updated with most current value

      
    }
  }
};
