//
//  JavaDump 1.0
//  Copyright (c) 1997 Matt T. Yourst. All rights reserved.
//
//  File:     DumpClassToHtml.java
//  Module:   DumpClassToHtml class
//  Version:  1.00.053
//  Date:     May - June 1997
//  Author:   Matt Yourst [yourst@laserstars.com]
//  Language: Java 1.1+
//
package lti.java.javadump;

import java.io.*;
import java.util.*;
import lti.java.jcf.*;

/**
 * Provides a method for documenting the low-level contents of
 * a class file (represented by a JcfClassFile object) in a
 * standard HTML file.
 */
public class DumpClassToHtml implements RuntimeConstants
{
  /**
   * If specified as part of a bit mask, each flag prevents
   * the dumping of the respective class file component.
   */
  public static final int OMIT_CONSTPOOL = 0x1;
  public static final int OMIT_CLASSDESC = 0x2;
  public static final int OMIT_INTERFACES = 0x4;
  public static final int OMIT_FIELDS = 0x8;
  public static final int OMIT_METHODS = 0x10;
  public static final int OMIT_EXTRAATTRIBUTES = 0x20;
  public static final int OMIT_ALLATTRIBUTES = 0x40;
  
  protected JcfClassFile jcf;
  protected JcfConstantPool cp;
  protected PrintStream output;
  protected int flags;
   
  /**
   * Creates a new DumpClassToHtml object with the specified source
   * class, output stream, and dumping flags.
   * @param flags Created as a bit mask of the OMIT_xxx fields of this class.
   */
  public DumpClassToHtml(JcfClassFile jcf, PrintStream output, int flags) throws Exception
  {
    super();
    this.jcf = jcf;
    this.cp = jcf.getConstantPool();
    this.output = output;
    this.flags = flags;
  }

  /**
   * Dump the class previously set up at instantiation to the output stream.
   */
  public void dump()
  {
    int i; // Utility variable
    int j; // Utility variable
    int m; // Maximum bound of utility variable
    Enumeration e; // Utility enumeration
    
    Date now = new Date(); // Get time of generation
    //
    // Prepare HTML header
    //
    println("<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 3.2//EN\">");
    println("<HTML><HEAD><TITLE>Class Dump: " + jcf.getFullName() + "</TITLE>");
    println("<META NAME=\"GENERATOR\" CONTENT=\"JavaDump 1.0 - Copyright 1997 Matt Yourst\">");
    println("</HEAD><BODY><P><FONT SIZE=+1>Dump of class:<BR></FONT>");
    println("<B><FONT SIZE=+2>" + jcf.getFullName().replace('/', '.') + "<BR></FONT>");
    println("</B><I>(Generated " + now.toString() + "</I><HR WIDTH=\"100%\"></P>");
    println("<FONT SIZE=+1><I><B>Table of Contents</B></I></FONT><BLOCKQUOTE><UL TYPE=DISC>");
    if ((flags & OMIT_CONSTPOOL) == 0) println("<LI><A HREF=\"#constpool\">Constant Pool</A></LI>");
    if ((flags & OMIT_CLASSDESC) == 0) println("<LI><A HREF=\"#classdesc\">Class Descriptor</A></LI>");
    if ((flags & OMIT_INTERFACES) == 0) println("<LI><A HREF=\"#interfaces\">Interfaces</A></LI>");
    if ((flags & OMIT_FIELDS) == 0) println("<LI><A HREF=\"#fields\">Fields</A></LI>");
    if ((flags & OMIT_METHODS) == 0) println("<LI><A HREF=\"#methods\">Methods</A></LI>");
    if ((flags & OMIT_EXTRAATTRIBUTES) == 0) println("<LI><A HREF=\"#attribs\">Attributes</A></LI>");
    println("</UL></BLOCKQUOTE>");
    //
    // Constant Pool
    //
    if ((flags & OMIT_CONSTPOOL) == 0)
    {
      println("<A NAME=\"constpool\">");
      cp = jcf.getConstantPool();
      printConstantPool();
    }
    //
    // Class Descriptor
    //
    if ((flags & OMIT_CLASSDESC) == 0)
    {
      println("<A NAME=\"classdesc\">");    
      println("<P><B><FONT SIZE=+1>Class Descriptor</FONT></B></P>");
      println("<BLOCKQUOTE><P><B>Access flags:</B> 0x" +
      Integer.toHexString(jcf.hdrAccessFlags) + "<BR>");
      println("<B>This class:</B> " + cpRef(jcf.hdrThisClassName) +
        " (" + cp.classNameAt(jcf.hdrThisClassName) + ") <BR>");
      println("<B>Superclass:</B> " + cpRef(jcf.hdrSuperClassName) +
        " (" + cp.classNameAt(jcf.hdrSuperClassName) + ") <BR>");
      println("</P></BLOCKQUOTE>");
    }
    //
    // Interfaces
    //
    if ((flags & OMIT_INTERFACES) == 0)
    {    
      println("<A NAME=\"interfaces\">");
      JcfInterfaceCollection interfaces = jcf.getInterfaces();
      m = interfaces.data.length;
      println("<P><FONT SIZE=+1><B>Interfaces</B></FONT> (" + m +
        ")</P><UL>");
      for (i = 0; i < m; i++)
      {
        j = interfaces.data[i];      
        println("<LI>" + cpRef(j) + " (" + cp.classNameAt(j) + ")</LI>");
      }
      println("</UL>");
    }
    //
    // Fields
    //
    JcfMemberCollection members;    
    if ((flags & OMIT_FIELDS) == 0)
    {
      println("<A NAME=\"fields\">");    
      members = jcf.getFields();
      printMembers(members, "Fields");
    }
    //
    // Methods
    //
    if ((flags & OMIT_METHODS) == 0)
    {
      println("<A NAME=\"methods\">");    
      members = jcf.getMethods();
      printMembers(members, "Methods");
    }
    //
    // Extra attributes
    //
    if ((flags & OMIT_EXTRAATTRIBUTES) == 0)
    {
      println("<A NAME=\"attribs\">");    
      println("<P><FONT SIZE=+1><B>Attributes</B> (" + jcf.getAttributes().size() + ")</FONT></P>");
      printAttributes(jcf.getAttributes());
    }
    println("<HR><I>End of report</I><BR>");
    println("Report produced by JavaDump 1.0. &copy; Copyright 1997 Matt Yourst.</BODY></HTML>");
  }

  /**
   * Returns an HTML hyperlink string (anchor) that has the text
   * "#123" (123 is in place of [cpr]) and links to the anchor #cp123.
   */
  protected static String cpRef(int cpr)
  {
    return "<A HREF=\"#cp" + cpr + "\">#" + cpr + "</A>";
  }
  
  /**
   * Print the constant pool (e.g. in a table format.)
   */
  protected void printConstantPool()
  {
    Enumeration e;
    int i;
    println("<P><B><FONT SIZE=+1>Constant Pool</FONT> (" + cp.size() + " slots)</B></P>");
    println("<BLOCKQUOTE><TABLE CELLSPACING=0 CELLPADDING=0>");
    println("<TR><TH ALIGN=LEFT VALIGN=TOP>Index</TH>");
    println("<TH ALIGN=LEFT VALIGN=TOP>Type</TH>");
    println("<TH ALIGN=LEFT VALIGN=TOP>Data</TH></TR>");
    e = cp.elements();
    i = 0;
    while (e.hasMoreElements())
    {
      CptGeneric cpEntry = (CptGeneric)e.nextElement();
      if (cpEntry == null)
      {
        i++;
        continue; // Nulls are possible after 64-bit (2 slot) tags
      }
      print("<TR><A NAME=\"cp" + i + "\">");
      print("<TD>" + i++ + "</TD><TD>");
      switch (cpEntry.getTag())
      {
        case CONSTANT_UTF8:
          print("UTF-8</TD><TD>\"" + ((CptUTF)cpEntry).value + "\" <I>(" +
            ((CptUTF)cpEntry).value.length() + " chars)</I>"); break;
        case CONSTANT_NAMEANDTYPE:
          print("NameType</TD><TD>name " + cpRef(((CptNameType)cpEntry).nameIndex) +
          " type " + cpRef(((CptNameType)cpEntry).signatureIndex)); break;
        case CONSTANT_CLASS:
          print("Class</TD><TD>name " + cpRef(((CptClass)cpEntry).nameIndex));
          break;
        case CONSTANT_METHOD:
          print("Method</TD><TD>class " +
          cpRef(((CptMemberOrInterface)cpEntry).classIndex) +
          " type " + cpRef(((CptMemberOrInterface)cpEntry).nameTypeIndex)); break;
        case CONSTANT_FIELD:
          print("Field</TD><TD>class " +
          cpRef(((CptMemberOrInterface)cpEntry).classIndex) +
          " type " + cpRef(((CptMemberOrInterface)cpEntry).nameTypeIndex)); break;
        case CONSTANT_STRING:
          print("String</TD><TD>data " + cpRef(((CptString)cpEntry).StringIndex));
          break;
        case CONSTANT_INTEGER:
          print("Integer</TD><TD>" + ((CptInteger)cpEntry).value);
          break;
        case CONSTANT_FLOAT:
          print("Float</TD><TD>" + ((CptFloat)cpEntry).value);
          break;
        case CONSTANT_LONG:
          print("Long</TD><TD>" + ((CptLong)cpEntry).value);
          break;
        case CONSTANT_DOUBLE:
          println("Double</TD><TD>" + ((CptDouble)cpEntry).value);
          break;
        case CONSTANT_INTERFACEMETHOD:
          println("IntMethod</TD><TD>class " +
          cpRef(((CptMemberOrInterface)cpEntry).classIndex) +
          " type " + cpRef(((CptMemberOrInterface)cpEntry).nameTypeIndex)); break;
        case CONSTANT_UNICODE:
          println("Unicode</TD><TD>" + ((CptUnicode)cpEntry).value.length + " bytes");
          break;
        default:
          println("???</TD><TD>???");
          break;
      }
      println("</TD></TR>");
    }
    println("</TABLE></BLOCKQUOTE>");
  }
  
  /**
   * Print descriptions of all members in the collection.
   * @param fieldOrMethod the text ("Fields" or "Methods") to print as a heading.
   */
  protected void printMembers(JcfMemberCollection members, String fieldOrMethod)
  {
    Enumeration e = members.elements();
    println("<P><B><FONT SIZE=+1>" + fieldOrMethod + "</FONT></B> (" +
      members.size() + ")</P><UL TYPE=SQUARE>");
    while (e.hasMoreElements())
    {
      println();
      JcfMember member = (JcfMember)e.nextElement();
      println("<LI><B>" + cp.utfAt(member.mbrNameIndex) + " </B>(" +
        cpRef(member.mbrNameIndex) + ") <I>signature</I> \"" +
        cp.utfAt(member.mbrSignatureIndex) + "\" (" +
        cpRef(member.mbrSignatureIndex) + ") <I>access</I> 0x" +
        Integer.toHexString(member.mbrAccessFlags) + "</LI>");
      printAttributes(member.getAttributes());
    }
    println("</UL>");
  }

  /**
   * Print an attribute collection. Should recurse through nested
   * attributes using appropriate indentation.
   */
  protected void printAttributes(JcfAttributeCollection attrs)
  {
    if ((flags & OMIT_ALLATTRIBUTES) != 0) return;
    println("<UL TYPE=DISC>");
    Enumeration e = attrs.elements();
    while (e.hasMoreElements())
    {
      byte[] data;
      JcfAttribute attr = (JcfAttribute)e.nextElement();
      String attrname = cp.utfAt(attr.getNameIndex());
      print("<LI><FONT COLOR=\"#FF0000\">&quot;<B>" + attrname + "</B>&quot;</FONT> (" +
        cpRef(attr.getNameIndex()) + "): " + attr.getLength() + " bytes");
      if (attrname.equals("Code"))
      {
        JcfCodeAttribute codeattr = (JcfCodeAttribute)attr;
        println(" [<I>stack words:</I> " + codeattr.catMaxStack +
          ", <I>local words:</I> " + codeattr.catMaxLocals +
          ", <I>bytecode:</I> " + codeattr.catCode.length + " bytes]");
        printAttributes(codeattr.catExtraAttributes);
      } else if (attrname.equals("ConstantValue")) {
        data = attr.getData();
        println(" [constant: " + cpRef((data[0] << 8) | data[1]) + "]");
      } else if (attrname.equals("SourceFile")) {
        data = attr.getData();
        int idx = ((data[0] & 0xff) << 8) | (data[1] & 0xff);
        println(" [filename: " + cpRef(idx) + " = " + cp.utfAt(idx) + "]");
      }
      println("</LI>");
    }
    println("</UL>");
  }

  /**
   * Print a line. Convenience method.
   */
  protected final void println(String s)
  {
    output.println(s);
  }
  /**
   * Print a text fragment. Convenience method.
   */
  protected final void print(String s)
  {
    output.print(s);
  }
  /**
   * Print an empty line. Convenience method.
   */
  protected final void println()
  {
    output.println();
  }
}
